@@ -7,6 +7,14 @@ set -o pipefail
 shopt -s failglob
 set -u
 
+log () {
+	echo "steam.sh[$$]: $*" >&2 || :
+}
+# This version interprets backslash escapes like echo -e
+log_e () {
+	echo -e "steam.sh[$$]: $*" >&2 || :
+}
+
 # Allow us to debug what's happening in the script if necessary
 if [ "${STEAM_DEBUG-}" ]; then
 	set -x
@@ -23,8 +31,8 @@ ARCHIVE_EXT=tar.xz
 # $PWD
 
 STEAMROOT="$(cd "$(dirname "$0")" && echo $PWD)"
-if [ -z ${STEAMROOT} ]; then
-	echo $"Couldn't find Steam root directory from "$0", aborting!"
+if [ -z "${STEAMROOT}" ]; then
+	log $"Couldn't find Steam root directory from "$0", aborting!"
 	exit 1
 fi
 STEAMDATA="$STEAMROOT"
@@ -33,11 +41,11 @@ if [ -z ${STEAMEXE-} ]; then
 fi
 # Backward compatibility for server operators
 if [ "$STEAMEXE" = "steamcmd" ]; then
-	echo "***************************************************"
-	echo "The recommended way to run steamcmd is: steamcmd.sh $*"
-	echo "***************************************************"
+	log "***************************************************"
+	log "The recommended way to run steamcmd is: steamcmd.sh $*"
+	log "***************************************************"
 	exec "$STEAMROOT/steamcmd.sh" "$@"
-	echo "Couldn't find steamcmd.sh" >&1
+	log "Couldn't find steamcmd.sh"
 	exit 255
 fi
 cd "$STEAMROOT"
@@ -71,8 +79,7 @@ function show_message()
 		;;
 	esac
 
-	# Show the message on standard output, for logging
-	echo -e "$title: $*"
+	log_e "$title: $*"
 
 	if [ -n "${STEAM_ZENITY}" ]; then
 		if ! "${STEAM_ZENITY}" "$style" --text="$*" --width=450 2>/dev/null; then
@@ -80,7 +87,7 @@ function show_message()
 			tmpfile="$(mktemp || echo "/tmp/steam_message.txt")"
 			echo -e "$*" >"$tmpfile"
 			xterm -bg "#383635" -fg "#d1cfcd" -T "$title" -e "cat $tmpfile; echo -n 'Press enter to continue: '; read input" 2>/dev/null || \
-				(echo "$title:"; cat "$tmpfile"; echo -n 'Press enter to continue: '; read input)
+				(echo "$title:"; cat "$tmpfile"; echo -n 'Press enter to continue: '; read input) >&2
 			rm -f "$tmpfile"
 		fi
 	else
@@ -152,10 +159,10 @@ function detect_platform()
 function detect_universe()
 {
 	if test -f "$STEAMROOT/Steam.cfg" && \
-		 egrep '^[Uu]niverse *= *[Bb]eta$' "$STEAMROOT/Steam.cfg" >/dev/null; then
+		grep -E '^[Uu]niverse *= *[Bb]eta$' "$STEAMROOT/Steam.cfg" >/dev/null; then
 		STEAMUNIVERSE="Beta"
 	elif test -f "$STEAMROOT/steam.cfg" && \
-		 egrep '^[Uu]niverse *= *[Bb]eta$' "$STEAMROOT/steam.cfg" >/dev/null; then
+		grep -E '^[Uu]niverse *= *[Bb]eta$' "$STEAMROOT/steam.cfg" >/dev/null; then
 		STEAMUNIVERSE="Beta"
 	else
 		STEAMUNIVERSE="Public"
@@ -214,7 +221,7 @@ function install_bootstrap()
 
 	STEAMBOOTSTRAPARCHIVE=`detect_bootstrap`
 	if [ -f "$STEAMBOOTSTRAPARCHIVE" ]; then
-		echo "Installing bootstrap $STEAMBOOTSTRAPARCHIVE"
+		log "Installing bootstrap $STEAMBOOTSTRAPARCHIVE"
 		tar xf "$STEAMBOOTSTRAPARCHIVE"
 		STATUS=$?
 	else
@@ -343,7 +350,7 @@ function unpack_runtime()
 	fi
 
 	# Make sure we haven't already unpacked them
-	if [ -f "$STEAM_RUNTIME/checksum" ] && cmp "$STEAM_RUNTIME.checksum" "$STEAM_RUNTIME/checksum" >/dev/null; then
+	if [ -f "$STEAM_RUNTIME/checksum" ] && [[ $(< "$STEAM_RUNTIME.checksum" ) = $(< "$STEAM_RUNTIME/checksum" ) ]] ; then
 		check_pins "$STEAM_RUNTIME"
 		return 0
 	fi
@@ -356,7 +363,7 @@ function unpack_runtime()
 	EXISTING_CHECKSUM="$(cd "$(dirname "$STEAM_RUNTIME")"; md5sum "$(basename "$STEAM_RUNTIME.$ARCHIVE_EXT")")"
 	EXPECTED_CHECKSUM="$(cat "$STEAM_RUNTIME.checksum")"
 	if [ "$EXISTING_CHECKSUM" != "$EXPECTED_CHECKSUM" ]; then
-		echo $"Runtime checksum: $EXISTING_CHECKSUM, expected $EXPECTED_CHECKSUM" >&2
+		log $"Runtime checksum: $EXISTING_CHECKSUM, expected $EXPECTED_CHECKSUM"
 		return 2
 	fi
 	if ! extract_archive $"Unpacking Steam Runtime" "$STEAM_RUNTIME.$ARCHIVE_EXT" "$EXTRACT_TMP"; then
@@ -409,7 +416,7 @@ function check_requirements()
 {
 	check_requirements_bin="$(command -v steam-runtime-check-requirements)"
 	if [ -z "$check_requirements_bin" ]; then
-		echo "Can't find 'steam-runtime-check-requirements', continuing anyway"
+		log "Can't find 'steam-runtime-check-requirements', continuing anyway"
 	else
 		message="$("$check_requirements_bin")"
 		status=$?
@@ -423,9 +430,9 @@ function check_requirements()
 			if [ -z "$message" ]; then
 				message="steam-runtime-check-requirements failed with an unexpected nonzero exit status. Trying to continue"
 			fi
-			echo $message
+			log "$message"
 		else
-			echo "Steam client's requirements are satisfied"
+			log "Steam client's requirements are satisfied"
 		fi
 	fi
 }
@@ -445,7 +452,7 @@ function reset_steam()
 
 	# Don't wipe development files
 	if [ -f "$STEAMROOT/steam_dev.cfg" ]; then
-		echo "Can't reset development directory"
+		log "Can't reset development directory"
 		return 1
 	fi
 
@@ -501,10 +508,10 @@ function reset_steam()
 
 		# Restore the steam data link
 		ln -s "$STEAMDATA" "$STEAMDATALINK"
-		echo $"Reset complete!"
+		log $"Reset complete!"
 	else
 		STATUS=1
-		echo $"Reset failed!"
+		log $"Reset failed!"
 	fi
 
 	# Okay, at this point we can recover, so re-enable interrupts
@@ -619,7 +626,7 @@ if [ "$INITIAL_LAUNCH" ]; then
 fi
 
 # Show what we detect for distribution and release
-echo "Running Steam on $(distro_description)"
+log "Running Steam on $(distro_description)"
 
 # The Steam runtime is a complete set of libraries for running
 # Steam games, and is intended to continue to work going forward.
@@ -651,7 +658,7 @@ if [ "$STEAM_RUNTIME" = "debug" ]; then
 
 			STEAM_RUNTIME_DEBUG_ARCHIVE="$STEAM_RUNTIME_DEBUG_DIR/$(basename "$STEAM_RUNTIME_DEBUG_URL")"
 			if [ ! -f "$STEAM_RUNTIME_DEBUG_ARCHIVE" ]; then
-				echo $"Downloading debug runtime: $STEAM_RUNTIME_DEBUG_URL"
+				log $"Downloading debug runtime: $STEAM_RUNTIME_DEBUG_URL"
 				(cd "$STEAM_RUNTIME_DEBUG_DIR" && \
 					download_archive $"Downloading debug runtime..." "$STEAM_RUNTIME_DEBUG_URL")
 			fi
@@ -660,48 +667,48 @@ if [ "$STEAM_RUNTIME" = "debug" ]; then
 			fi
 		fi
 		if [ -d "$STEAM_RUNTIME_DEBUG_DIR/$STEAM_RUNTIME_DEBUG" ]; then
-			echo "STEAM_RUNTIME debug enabled, using $STEAM_RUNTIME_DEBUG"
+			log "STEAM_RUNTIME debug enabled, using $STEAM_RUNTIME_DEBUG"
 			export STEAM_RUNTIME="$STEAM_RUNTIME_DEBUG_DIR/$STEAM_RUNTIME_DEBUG"
 
 			# Set up the link to the source code
 			ln -sf "$STEAM_RUNTIME/source" /tmp/source
 		else
-			echo $"STEAM_RUNTIME couldn't download and unpack $STEAM_RUNTIME_DEBUG_URL, falling back to $STEAM_RUNTIME"
+			log $"STEAM_RUNTIME couldn't download and unpack $STEAM_RUNTIME_DEBUG_URL, falling back to $STEAM_RUNTIME"
 		fi
 	fi
 elif [ "$STEAM_RUNTIME" = "1" ]; then
-	echo "STEAM_RUNTIME is enabled by the user"
+	log "STEAM_RUNTIME is enabled by the user"
 	export STEAM_RUNTIME="$STEAMROOT/$PLATFORM/steam-runtime"
 elif [ "$STEAM_RUNTIME" = "0" ]; then
-	echo "STEAM_RUNTIME is disabled by the user"
+	log "STEAM_RUNTIME is disabled by the user"
 elif [ -z "$STEAM_RUNTIME" ]; then
 	if runtime_supported; then
-		echo "STEAM_RUNTIME is enabled automatically"
+		log "STEAM_RUNTIME is enabled automatically"
 		export STEAM_RUNTIME="$STEAMROOT/$PLATFORM/steam-runtime"
 	else
-		echo "STEAM_RUNTIME is disabled automatically"
+		log "STEAM_RUNTIME is disabled automatically"
 	fi
 else
-	echo "STEAM_RUNTIME has been set by the user to: $STEAM_RUNTIME"
+	log "STEAM_RUNTIME has been set by the user to: $STEAM_RUNTIME"
 fi
 if [ "$STEAM_RUNTIME" -a "$STEAM_RUNTIME" != "0" ]; then
 	# Unpack the runtime if necessary
 	if unpack_runtime; then
 		if [ ! -x "$STEAM_RUNTIME/setup.sh" ]; then
-			echo "internal error: $STEAM_RUNTIME/setup.sh is missing, this runtime is invalid or corrupted"
+			log "internal error: $STEAM_RUNTIME/setup.sh is missing, this runtime is invalid or corrupted"
 			exit 1
 		fi
 		export PATH="$("$STEAM_RUNTIME/setup.sh" --print-bin-path):$PATH"
 
 		if [ ! -x "$STEAM_RUNTIME/run.sh" ]; then
-			echo "internal error: $STEAM_RUNTIME/run.sh is missing, this runtime is invalid or corrupted"
+			log "internal error: $STEAM_RUNTIME/run.sh is missing, this runtime is invalid or corrupted"
 			exit 1
 		fi
 		export STEAM_RUNTIME_LIBRARY_PATH="$("$STEAM_RUNTIME/run.sh" --print-steam-runtime-library-paths)"
 
 		export LD_LIBRARY_PATH="$STEAM_RUNTIME_LIBRARY_PATH"
 	else
-		echo "Unpack runtime failed, error code $?"
+		log "Unpack runtime failed, error code $?"
 		show_message --error $"Couldn't set up the Steam Runtime. Are you running low on disk space?\nContinuing..."
 	fi
 fi
@@ -710,7 +717,7 @@ if [ "${1-}" = "--run" ]; then
 	STEAM_RUNTIME_RUN_SCRIPT="$STEAM_RUNTIME/run.sh"
 	shift
 	exec "$STEAM_RUNTIME/run.sh" "$@"
-	echo "Couldn't find run.sh" >&1
+	log "Couldn't find run.sh"
 	exit 255
 fi
 
@@ -761,8 +768,8 @@ if [ "$STEAM_DEBUGGER" == "gdb" ] || [ "$STEAM_DEBUGGER" == "cgdb" ]; then
 	fi
 
 	# Enable index file caching for reasonable launch time under gdb
-	echo set index-cache on >> "$ARGSFILE"
-	echo show index-cache >> "$ARGSFILE"
+	echo set index-cache enabled on >> "$ARGSFILE"
+	echo show index-cache stats >> "$ARGSFILE"
 
 	: "${DEBUGGER_ARGS=}"
 	echo "gdb/cgdb launch with ARGSFILE: $ARGSFILE"
@@ -777,10 +784,16 @@ elif [ "$STEAM_DEBUGGER" == "callgrind" ]; then
 	valgrind --tool=callgrind --instr-atstart=no "$STEAMROOT/$STEAMEXEPATH" "$@"
 elif [ "$STEAM_DEBUGGER" == "strace" ]; then
 	strace -osteam.strace "$STEAMROOT/$STEAMEXEPATH" "$@"
+elif [ "$STEAM_DEBUGGER" == "lldb-mi" ]; then
+	$STEAM_DEBUGGER $DEBUGGER_ARGS "$STEAMROOT/$STEAMEXEPATH" -- "$@"
+elif [ "$STEAM_DEBUGGER" == "gdbserver" ]; then
+	$STEAM_DEBUGGER $DEBUGGER_ARGS "$STEAMROOT/$STEAMEXEPATH" "$@"
+elif [ -z "$STEAM_DEBUGGER" ]; then
+	"$STEAMROOT/$STEAMEXEPATH" "$@"
 else
 	# Most likely not what you want!
-	echo "WARNING: Using default/fallback debugger launch"
-	echo $STEAM_DEBUGGER $DEBUGGER_ARGS "$STEAMROOT/$STEAMEXEPATH" "$@"
+	log "WARNING: Using default/fallback debugger launch"
+	echo $STEAM_DEBUGGER $DEBUGGER_ARGS "$STEAMROOT/$STEAMEXEPATH" "$@" >&2
 	$STEAM_DEBUGGER $DEBUGGER_ARGS "$STEAMROOT/$STEAMEXEPATH" "$@"
 fi
 STATUS=$?
@@ -791,6 +804,6 @@ export LD_LIBRARY_PATH="$SYSTEM_LD_LIBRARY_PATH"
 
 # If steam requested to restart, then restart
 if [ $STATUS -eq $MAGIC_RESTART_EXITCODE ] ; then
-	echo "Restarting Steam by request..."
+	log "Restarting Steam by request..."
 	exec "$0" "$@"
 fi
